# Generated by Django 5.1.6 on 2025-02-24 12:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('history4feed', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='feed',
            name='freshness',
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name='feed',
            name='feed_type',
            field=models.CharField(choices=[('rss', 'Rss'), ('atom', 'Atom'), ('skeleton', 'Skeleton'), ('search_index', 'Search Index')], editable=False, help_text='type of feed', max_length=12),
        ),
    ]



import os
from celery import Celery
# Set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'history4feed.settings')

app = Celery('history4feed')


app.config_from_object('os:environ', namespace='CELERY')

# Load task modules from all registered Django apps.
app.autodiscover_tasks()


from unittest.mock import MagicMock, patch
import uuid

import pytest
from history4feed.app.models import Feed, FeedType, Job
from history4feed.app.views import FeedView
import pytest
from rest_framework import status
from django.http import HttpRequest
from rest_framework.request import Request
from history4feed.app.models import Post, Feed, Job
from datetime import datetime as dt
from unittest.mock import patch
from history4feed.app.serializers import (
    FeedFetchSerializer,
    FeedPatchSerializer,
    FeedSerializer,
    PostWithFeedIDSerializer,
    SearchIndexFeedSerializer,
)
from history4feed.app.views import PostOnlyView
from dateutil.parser import parse as parse_date


from history4feed.app.utils import (
    DatetimeFieldUTC,
    Ordering,
    Pagination,
    MinMaxDateFilter,
)

# from .openapi_params import FEED_PARAMS, POST_PARAMS

from history4feed.app.serializers import PostWithFeedIDSerializer
from history4feed.app.models import Post, Feed, Job
from rest_framework import (
    status,
)
from django_filters.rest_framework import (
    DjangoFilterBackend,
)


def test_class_variables():
    assert FeedView.ordering_fields == [
        "datetime_added",
        "title",
        "url",
        "count_of_posts",
        "earliest_item_pubdate",
        "latest_item_pubdate",
    ]
    assert FeedView.ordering == "datetime_added_descending", "default ordering"
    assert FeedView.filter_backends == [
        DjangoFilterBackend,
        Ordering,
        MinMaxDateFilter,
    ]
    assert type(FeedView.pagination_class) == Pagination
    assert FeedView.pagination_class.results_key == "feeds"
    assert FeedView.serializer_class == FeedSerializer


@pytest.mark.django_db
def test_create_feed(client):

    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")
    job = Job.objects.create(feed=feed)
    payload = dict(data="some data")
    with patch.object(FeedView, "new_create_job") as mock_create_job:
        mock_create_job.return_value = job
        resp = client.post(
            "/api/v1/feeds/", data=payload, content_type="application/json"
        )
        assert resp.status_code == 201
        assert str(resp.data["job_id"]) == str(job.id)


@pytest.mark.parametrize("include_remote_blogs", [True, False])
@pytest.mark.parametrize("use_search_index", [True, False])
@pytest.mark.django_db
def test_new_create_job(use_search_index, include_remote_blogs):
    request = MagicMock()
    request.data = dict(
        title="some title",
        description="some description",
        url="https://example.net/",
        use_search_index=use_search_index,
        include_remote_blogs=include_remote_blogs,
    )
    with (
        patch(
            "history4feed.app.views.task_helper.new_job",
            side_effect=lambda *args: Job.objects.create(feed=args[0]),
        ) as mock_new_job,
        patch(
            "history4feed.app.views.SearchIndexFeedSerializer",
            side_effect=SearchIndexFeedSerializer,
        ) as mock_search_index_serializer,
        patch(
            "history4feed.app.views.h4f.parse_feed_from_url"
        ) as mock_parse_feed_from_url,
    ):
        mock_parse_feed_from_url.return_value = {**request.data, "feed_type": "atom"}
        result = FeedView().new_create_job(request)
        assert result.feed.title == request.data["title"]
        assert result.feed.description == request.data["description"]
        mock_new_job.assert_called_once_with(result.feed, include_remote_blogs)
        if use_search_index:
            mock_search_index_serializer.assert_called_once_with(data=request.data)
            assert result.feed.feed_type == FeedType.SEARCH_INDEX
            mock_parse_feed_from_url.assert_not_called()
        else:
            mock_parse_feed_from_url.assert_called_once_with(request.data["url"])
            assert result.feed.feed_type == FeedType.ATOM
            mock_search_index_serializer.assert_not_called()


@pytest.mark.django_db
def test_create_skeleton(client):
    payload = dict(
        title="some title", description="some description", url="https://example.net/"
    )
    resp = client.post(
        "/api/v1/feeds/skeleton/", data=payload, content_type="application/json"
    )
    assert resp.status_code == 201, resp.content
    assert resp.data["title"] == "some title"
    assert resp.data["description"] == "some description"
    assert resp.data["feed_type"] == "skeleton"


@pytest.mark.django_db
def test_create_skeleton_400(client):
    payload = dict(title="some title", description="some description")
    resp = client.post(
        "/api/v1/feeds/skeleton/", data=payload, content_type="application/json"
    )
    assert resp.status_code == 400, resp.content


@pytest.mark.parametrize(
    "payload",
    [
        dict(title="my new title"),
        dict(description="my new descr"),
        dict(description="my new descr", title="new title 2"),
        dict(pretty_url="https://aa.net/3", title="new title 2"),
        dict(pretty_url="https://example.com/4", description="new descr 2"),
    ],
)
@pytest.mark.django_db
def test_feed_metadata_update(client, payload):
    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")
    with patch(
        "history4feed.app.views.FeedPatchSerializer",
        side_effect=FeedPatchSerializer,
    ) as mock_patch_feed_serializer:
        resp = client.patch(
            f"/api/v1/feeds/{feed.id}/", data=payload, content_type="application/json"
        )
        assert resp.status_code == 201, resp.content
        mock_patch_feed_serializer.assert_called_once_with(
            feed, data=payload, partial=True
        )
        for k in ["title", "description", "pretty_url"]:
            if k in payload:
                assert resp.data[k] == payload[k]


@pytest.mark.django_db
def test_fetch_feed(client):

    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")
    job = Job.objects.create(feed=feed)
    with patch.object(FeedView, "new_fetch_job") as mock_fetch_job:
        mock_fetch_job.return_value = job
        url = f"/api/v1/feeds/{feed.id}/fetch/"
        resp = client.patch(url, content_type="application/json")
        assert resp.status_code == 201, resp.content
        assert str(resp.data["job_id"]) == str(job.id)
        mock_fetch_job.assert_called_once()
        mock_fetch_job.call_args[0][0].path == url


@pytest.mark.django_db
def test_new_fetch_job():
    request = Request(HttpRequest())
    feed = Feed.objects.create(
        url="https://example.com/rss.xml", title="Test Feed", feed_type="atom"
    )
    view = FeedView()
    view.request = request
    view.kwargs = dict(feed_id=feed.id)
    with (
        patch(
            "history4feed.app.views.task_helper.new_job",
            side_effect=lambda *args: Job.objects.create(feed=args[0]),
        ) as mock_new_job,
        patch(
            "history4feed.app.views.FeedFetchSerializer",
            side_effect=FeedFetchSerializer,
        ) as mock_fetch_feed_serializer,
    ):
        result = view.new_fetch_job(request)
        assert result.feed == feed
        mock_new_job.assert_called_once_with(result.feed, False)
        mock_fetch_feed_serializer.assert_called_once_with(
            feed, data=request.data, partial=True
        )


@pytest.mark.parametrize(
    ["filters", "expected"],
    [
        [dict(), (0, 1, 2)],
        [dict(title="fEed"), (0, 1, 2)],
        [dict(title="feed 2"), (1,)],
        [dict(description="-iption", title="feed"), (1,)],
        [dict(feed_type="skeleton"), (2,)],
    ],
)
@pytest.mark.django_db
def test_list_feed(client, filters, expected):
    feeds = [
        Feed.objects.create(
            url="https://example.com/rss1.xml",
            title="Test Feed 1",
            feed_type="atom",
            description="some description",
        ),
        Feed.objects.create(
            url="https://example.com/rss2.xml",
            title="Test Feed 2",
            feed_type="rss",
            description="descr-iption",
        ),
        Feed.objects.create(
            url="https://example.com/rss3.xml",
            title="Some other feed 3",
            feed_type="skeleton",
        ),
    ]
    expected_ids = {str(feeds[i].id) for i in expected}
    resp = client.get("/api/v1/feeds/", query_params=filters)
    assert resp.status_code == 200
    assert len(resp.data["feeds"]) == len(expected_ids)
    assert {feed["id"] for feed in resp.data["feeds"]} == expected_ids


@pytest.mark.django_db
def test_retrieve(client):
    feed = Feed.objects.create(
        url="https://example.com/rss1.xml",
        title="Test Feed 1",
        feed_type="atom",
        description="some description",
    )
    another_uuid = uuid.uuid4()
    resp = client.get(f"/api/v1/feeds/{feed.id}/")
    assert resp.status_code == 200
    assert resp.data["id"] == str(feed.id)

    resp = client.get(f"/api/v1/feeds/{another_uuid}/")
    assert resp.status_code == 404





import os
import time
from types import SimpleNamespace
import unittest, pytest
from urllib.parse import urljoin

from tests.utils import remove_unknown_keys, wait_for_jobs

base_url = os.environ["SERVICE_BASE_URL"]
import requests

@pytest.mark.parametrize(
    ["feed_id", "should_fail"],
    [
        ["c2fe0594-f463-5362-afe7-6950bda94bc6", True], #feed does not exist
        ["9c04d319-a949-52df-bcb6-5a73a1458fe5", False],
        ["9c04d319-a949-52df-bcb6-5a73a1458fe5", True], #feed already deleted
    ]
)
def test_delete_feed(feed_id, should_fail):
    feed_url = urljoin(base_url, f"api/v1/feeds/{feed_id}/")
    delete_resp = requests.delete(feed_url)

    if should_fail:
        assert delete_resp.status_code == 404, f"delete feed request expected to fail: {delete_resp.text}"
        return
    assert delete_resp.status_code == 204, f"unexpected status, body: {delete_resp.text}"


    get_resp = requests.get(feed_url)
    assert get_resp.status_code == 404, f"feed should already be deleted"



from unittest.mock import MagicMock, patch, call

import celery
import celery.canvas
import pytest
from history4feed.app import models
from history4feed.app.models import Feed, FeedType, FullTextState, FulltextJob, Job, Post
from history4feed.h4fscripts import exceptions
from history4feed.h4fscripts.h4f import PostDict
from history4feed.h4fscripts.task_helper import (
    JobCancelled,
    add_post_to_db,
    create_fulltexts_task_chain,
    new_job,
    new_patch_posts_job,
    retrieve_full_text,
    retrieve_posts_from_links,
    retrieve_posts_from_serper,
    retrieve_posts_from_url,
    start_job,
    start_post_job,
)
from rest_framework.exceptions import APIException, Throttled
from datetime import UTC, datetime as dt


@pytest.fixture(autouse=True, scope="module")
def celery_eager():
    from history4feed.h4fscripts.celery import app

    app.conf.task_always_eager = True
    app.conf.broker_url = None
    yield
    app.conf.task_always_eager = False


@pytest.mark.django_db
def test_new_job():
    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")

    with (
        patch(
            "history4feed.h4fscripts.task_helper.retrieve_posts_from_links.run"
        ) as mock_retrieve_posts,
        patch("history4feed.h4fscripts.task_helper.start_job.run") as mock_start_job,
        patch(
            "history4feed.h4fscripts.task_helper.collect_and_schedule_removal.run"
        ) as mock_collect_and_schedule_removal,
        patch(
            "history4feed.h4fscripts.task_helper.error_handler.run"
        ) as mock_error_handler,
        patch(
            "history4feed.h4fscripts.task_helper.queue_lock", return_value=True
        ) as mock_queue_lock,
    ):
        new_job(feed, include_remote_blogs=False)
        job: Job = mock_queue_lock.call_args[0][1]
        assert job.feed == feed
        mock_queue_lock.assert_called_once_with(feed, job)
        mock_retrieve_posts.assert_called_once_with(mock_start_job.return_value, job.id)
        mock_start_job.assert_called_once_with(job.id)
        mock_collect_and_schedule_removal.assert_called_once_with(job.id)
        mock_error_handler.assert_not_called(), "should not be called unless there is an error"


@pytest.mark.django_db
def test_job_queue_fail():
    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")

    with (
        patch(
            "history4feed.h4fscripts.task_helper.cache.get", return_value={}
        ) as mock_get_cache,
        patch(
            "history4feed.h4fscripts.task_helper.queue_lock", return_value=False
        ) as mock_queue_lock,
        pytest.raises(Throttled),
    ):
        new_job(feed, include_remote_blogs=False)
        job: Job = mock_queue_lock.call_args[0][1]


@pytest.mark.django_db
def test_new_patch_posts_job():
    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")
    posts = [
        Post.objects.create(feed=feed, title="First post", pubdate=dt.now()),
        Post.objects.create(
            feed=feed,
            title="Second post",
            pubdate=dt.now(),
            link="https://example.net/post2",
        ),
    ]

    with (
        patch(
            "history4feed.h4fscripts.task_helper.retrieve_full_text.run"
        ) as mock_retrieve_full_text,
        # patch("history4feed.h4fscripts.task_helper.start_post_job.run") as mock_start_post_job,
        patch(
            "history4feed.h4fscripts.task_helper.collect_and_schedule_removal.run"
        ) as mock_collect_and_schedule_removal,
        patch(
            "history4feed.h4fscripts.task_helper.error_handler.run"
        ) as mock_error_handler,
        patch(
            "history4feed.h4fscripts.task_helper.queue_lock", return_value=True
        ) as mock_queue_lock,
    ):
        job = new_patch_posts_job(feed, posts, include_remote_blogs=False)
        assert job.feed == feed
        mock_queue_lock.assert_called_once_with(feed, job)
        # assert job.state == models.JobState.RUNNING

        [
            models.FulltextJob.objects.get(pk=call[0][0]).post.id
            for call in mock_retrieve_full_text.call_args_list
        ] == [post.id for post in posts]

        mock_collect_and_schedule_removal.assert_called_once_with(job.id)
        mock_error_handler.assert_not_called(), "should not be called unless there is an error"


@pytest.mark.django_db
def test_start_post_job_already_cancelled():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed"),
        state=models.JobState.PENDING,
    )
    job_obj.update_state(models.JobState.CANCELLED)
    with (
        patch(
            "history4feed.h4fscripts.task_helper.queue_lock", return_value=True
        ) as mock_queue_lock,
    ):
        result = start_post_job.si(job_obj.id).delay()
        assert result.get() == False
        mock_queue_lock.assert_not_called()


@pytest.mark.django_db
def test_start_post_job_retries_until_queue_no_longer_locked():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed"),
        state=models.JobState.PENDING,
    )
    with (
        patch(
            "history4feed.h4fscripts.task_helper.queue_lock",
            side_effect=[False, False, True],
        ) as mock_queue_lock,
    ):
        result = start_post_job.si(job_obj.id).delay()
        job_obj.refresh_from_db()
        assert result.get() == True
        assert job_obj.state == models.JobState.RUNNING
        mock_queue_lock.assert_called()
        assert len(mock_queue_lock.call_args_list) == 3


@pytest.mark.django_db
def test_start_job__search_index():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(
            url="https://example.com/rss.xml",
            title="Test Feed",
            feed_type=models.FeedType.SEARCH_INDEX,
        ),
        state=models.JobState.PENDING,
    )
    result = start_job(job_obj.id)
    assert result == [job_obj.feed.url]
    job_obj.refresh_from_db()
    assert job_obj.state == models.JobState.RUNNING


@pytest.mark.django_db
def test_start_job__atom_or_rss():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(
            url="https://example.com/rss.xml",
            title="Test Feed",
            feed_type=models.FeedType.RSS,
        ),
        state=models.JobState.PENDING,
    )
    with patch(
        "history4feed.h4fscripts.wayback_helpers.get_wayback_urls"
    ) as mock_get_wayback_urls:
        result = start_job(job_obj.id)
        assert result == mock_get_wayback_urls.return_value
        job_obj.refresh_from_db()
        assert job_obj.state == models.JobState.RUNNING


@pytest.mark.django_db
def test_start_job__atom_or_rss__fails():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(
            url="https://example.com/rss.xml",
            title="Test Feed",
            feed_type=models.FeedType.RSS,
        ),
        state=models.JobState.PENDING,
    )
    with patch(
        "history4feed.h4fscripts.wayback_helpers.get_wayback_urls",
        side_effect=Exception,
    ) as mock_get_wayback_urls:
        result = start_job(job_obj.id)
        assert result == []
        job_obj.refresh_from_db()
        assert job_obj.state == models.JobState.FAILED


@pytest.mark.parametrize(
    "feed_type",
    [
        models.FeedType.RSS,
        models.FeedType.ATOM,
        models.FeedType.SEARCH_INDEX,
    ],
)
@pytest.mark.django_db
def test_retrieve_posts_from_links(feed_type):
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(
            url="https://example.com/rss.xml",
            title="Test Feed",
            feed_type=feed_type,
        ),
        state=models.JobState.PENDING,
    )
    chains = [MagicMock(), MagicMock()]
    with (
        patch(
            "history4feed.h4fscripts.task_helper.retrieve_posts_from_serper",
            return_value=MagicMock(),
        ) as mock_retrieve_serp,
        patch(
            "history4feed.h4fscripts.task_helper.retrieve_posts_from_url",
            return_value=[{}, MagicMock(), None],
        ) as mock_retrieve_rss,
        patch(
            "history4feed.h4fscripts.task_helper.create_fulltexts_task_chain",
            side_effect=chains,
        ) as mock_ft_chain,
    ):

        resp = retrieve_posts_from_links(
            ["https://goo.gl", "http://example.com"], job_obj.id
        )
        job_obj.refresh_from_db()
        assert job_obj.feed.freshness == job_obj.run_datetime
        if job_obj.feed.feed_type == models.FeedType.SEARCH_INDEX:
            mock_retrieve_serp.assert_has_calls(
                [
                    call(job_obj.feed, job_obj, "https://goo.gl"),
                    call(job_obj.feed, job_obj, "http://example.com"),
                ]
            )
            mock_ft_chain.call_count == 2
            mock_ft_chain.assert_called_with(
                job_obj.id, mock_retrieve_serp.return_value
            )
        else:
            mock_retrieve_rss.assert_has_calls(
                [
                    call(
                        "https://goo.gl",
                        job_obj.feed,
                        job_obj,
                    ),
                    call("http://example.com", job_obj.feed, job_obj),
                ]
            )
            mock_ft_chain.call_count == 2
            mock_ft_chain.assert_called_with(
                job_obj.id, mock_retrieve_rss.return_value[1]
            )
        assert resp == [chain.apply_async.return_value.id for chain in chains]


@pytest.mark.django_db
def test_create_fulltexts_task_chain():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(
            url="https://example.com/rss.xml",
            title="Test Feed",
            feed_type=models.FeedType.RSS,
        ),
        state=models.JobState.PENDING,
    )
    feed = job_obj.feed
    posts = [
        Post.objects.create(feed=feed, title="First post", pubdate=dt.now()),
        Post.objects.create(
            feed=feed,
            title="Second post",
            pubdate=dt.now(),
            link="https://example.net/post2",
        ),
    ]

    with (
        patch(
            "history4feed.h4fscripts.task_helper.retrieve_full_text.run"
        ) as mock_retrieve_ft,
    ):
        chain = create_fulltexts_task_chain(job_obj.id, posts)
        chain.apply_async()
        assert isinstance(chain, (celery.chain, celery.canvas._chain)), type(chain)
        fts = models.FulltextJob.objects.filter(job_id=job_obj.id)
        mock_retrieve_ft.assert_has_calls([call(ft.pk) for ft in fts], any_order=True)


@pytest.mark.django_db
def test_retrieve_posts_from_serper():
    job_obj = models.Job.objects.create(
        feed=Feed.objects.create(
            url="https://example.com/rss.xml",
            title="Test Feed",
            feed_type=models.FeedType.SEARCH_INDEX,
            freshness=dt(2024, 1, 1),
        ),
        state=models.JobState.PENDING,
    )
    with (
        patch(
            "history4feed.h4fscripts.task_helper.fetch_posts_links_with_serper",
            return_value={
                post.link: post
                for post in [
                    PostDict(
                        title="random title",
                        link="https://example.com/post/1",
                        pubdate=dt.now(),
                    ),
                    PostDict(
                        title="random title",
                        link="https://example.com/post/2",
                        pubdate=dt.now()
                    ),
                    PostDict(
                        title="random title",
                        link="https://example.com/post/3",
                        pubdate=dt.now()
                    ),
                    PostDict(
                        title="random title",
                        link="https://example.com/post/4",
                        pubdate=dt.now()
                    ),
                ]
            },
        ) as mock_fetch,
        patch(
            "history4feed.h4fscripts.task_helper.add_post_to_db",
            side_effect=add_post_to_db,
        ) as mock_add_post_to_db,
    ):
        url = "https://example.net"
        results = retrieve_posts_from_serper(job_obj.feed, job_obj, url)
        mock_fetch.assert_called_once_with(
            url, from_time=dt(2024, 1, 1, tzinfo=UTC), to_time=job_obj.run_datetime
        )
        assert len(results) == 4
        mock_add_post_to_db.call_count == 4

def test_retrieve_posts_from_url():
    pass




@pytest.fixture
def dummy_feed():
    return MagicMock(spec=Feed)


@pytest.fixture
def dummy_job():
    job = MagicMock(spec=Job)
    job.is_cancelled.return_value = False
    return job


@patch("history4feed.h4fscripts.task_helper.add_post_to_db")
@patch("history4feed.h4fscripts.task_helper.h4f.parse_posts_from_rss_feed")
@patch("history4feed.h4fscripts.task_helper.h4f.parse_feed_from_content")
@patch("history4feed.h4fscripts.task_helper.h4f.fetch_page_with_retries")
@patch("history4feed.h4fscripts.task_helper.time.sleep")
def test_retrieve_posts_rss_success(mock_sleep, mock_fetch, mock_parse_feed, mock_parse_rss, mock_add_post, dummy_feed, dummy_job):
    url = "https://example.com/rss"

    mock_fetch.return_value = (b"<xml>RSS</xml>", "text/xml", url)
    mock_parse_feed.return_value = {"feed_type": FeedType.RSS}
    mock_parse_rss.return_value = {"1": {"title": "Post 1"}, "2": {"title": "Post 2"}}
    post1 = MagicMock(spec=Post)
    post2 = MagicMock(spec=Post)
    mock_add_post.side_effect = [post1, post2]

    parsed_feed, all_posts, error = retrieve_posts_from_url(url, dummy_feed, dummy_job)

    assert parsed_feed['feed_type'] == FeedType.RSS
    assert all_posts == [post1, post2]
    assert error is None
    mock_fetch.assert_called_once_with(url)
    dummy_feed.save.assert_called_once()


@patch("history4feed.h4fscripts.task_helper.add_post_to_db")
@patch("history4feed.h4fscripts.task_helper.h4f.parse_posts_from_atom_feed")
@patch("history4feed.h4fscripts.task_helper.h4f.parse_feed_from_content")
@patch("history4feed.h4fscripts.task_helper.h4f.fetch_page_with_retries")
@patch("time.sleep")
def test_retrieve_posts_atom_success(mock_sleep, mock_fetch, mock_parse_feed, mock_parse_atom, mock_add_post, dummy_feed, dummy_job):
    url = "https://example.com/atom"
    mock_fetch.return_value = (b"<xml>ATOM</xml>", "text/xml", url)
    mock_parse_feed.return_value = {"feed_type": FeedType.ATOM}
    mock_parse_atom.return_value = {"1": {"title": "A1"}}
    post = MagicMock(spec=Post)
    mock_add_post.return_value = post

    parsed_feed, all_posts, error = retrieve_posts_from_url(url, dummy_feed, dummy_job)

    assert parsed_feed['feed_type'] == FeedType.ATOM
    assert all_posts == [post]
    assert error is None


@patch("history4feed.h4fscripts.task_helper.h4f.fetch_page_with_retries")
@patch("history4feed.h4fscripts.task_helper.h4f.parse_feed_from_content")
@patch("time.sleep")
def test_retrieve_posts_unknown_feedtype(mock_sleep, mock_parse_feed, mock_fetch, dummy_feed, dummy_job):
    url = "https://unknown.feed"
    mock_fetch.return_value = (b"<xml>???</xml>", "text/xml", url)
    mock_parse_feed.return_value = {"feed_type": "UNKNOWN"}

    _, _, err = retrieve_posts_from_url(url, dummy_feed, dummy_job)
    assert isinstance(err,  exceptions.UnknownFeedtypeException)


@patch("history4feed.h4fscripts.task_helper.h4f.fetch_page_with_retries")
def test_retrieve_posts_job_cancelled(mock_fetch, dummy_feed, dummy_job):
    dummy_job.is_cancelled.return_value = True
    url = "https://cancel.test"

    _, _, err =retrieve_posts_from_url(url, dummy_feed, dummy_job)
    assert isinstance(err,  JobCancelled)


@patch("history4feed.h4fscripts.task_helper.h4f.fetch_page_with_retries", side_effect=ConnectionError("Timeout"))
@patch("time.sleep")
def test_retrieve_posts_connection_error_retries(mock_sleep, mock_fetch, dummy_feed, dummy_job, settings):
    settings.REQUEST_RETRY_COUNT = 3
    settings.WAYBACK_SLEEP_SECONDS = 1

    url = "https://retry.fail"

    parsed_feed, all_posts, error = retrieve_posts_from_url(url, dummy_feed, dummy_job)

    assert parsed_feed == {}
    assert all_posts == []
    assert isinstance(error, ConnectionError)
    assert mock_fetch.call_count == 3
    assert mock_sleep.call_count == 2  # called only after the first attempt fails


@patch("history4feed.h4fscripts.task_helper.h4f.fetch_page_with_retries", side_effect=ValueError("boom"))
@patch("time.sleep")
def test_retrieve_posts_general_exception(mock_sleep, mock_fetch, dummy_feed, dummy_job):
    url = "https://explode.me"
    parsed_feed, all_posts, error = retrieve_posts_from_url(url, dummy_feed, dummy_job)

    assert parsed_feed == {}
    assert all_posts == []
    assert isinstance(error, ValueError)
    assert mock_fetch.call_count == 1


@pytest.fixture
def dummy_ftjob(dummy_post):
    ftjob = MagicMock(spec=FulltextJob)
    ftjob.post = dummy_post
    ftjob.pk = 1
    ftjob.status = None
    ftjob.error_str = ""
    return ftjob
@pytest.fixture
def dummy_post():
    post = MagicMock(spec=Post)
    post.link = "https://example.com/article"
    return post

@patch("history4feed.h4fscripts.task_helper.logger")
@patch("history4feed.h4fscripts.task_helper.h4f.get_full_text")
@patch("history4feed.h4fscripts.task_helper.models.FulltextJob.objects.get")
def test_retrieve_full_text_success(mock_get_job, mock_get_full_text, mock_logger, dummy_ftjob):
    mock_get_job.return_value = dummy_ftjob
    dummy_ftjob.is_cancelled.return_value = False
    mock_get_full_text.return_value = ("<p>Content</p>", "text/html")

    retrieve_full_text(dummy_ftjob.pk)

    mock_get_full_text.assert_called_once_with(dummy_ftjob.post.link)
    assert dummy_ftjob.status == FullTextState.RETRIEVED
    assert dummy_ftjob.error_str == ""
    assert dummy_ftjob.post.description == "<p>Content</p>"
    assert dummy_ftjob.post.content_type == "text/html"
    assert dummy_ftjob.post.is_full_text is True
    dummy_ftjob.save.assert_called_once()
    dummy_ftjob.post.save.assert_called_once()


@patch("history4feed.h4fscripts.task_helper.models.FulltextJob.objects.get")
def test_retrieve_full_text_cancelled(mock_get_job, dummy_ftjob):
    mock_get_job.return_value = dummy_ftjob
    dummy_ftjob.is_cancelled.return_value = True

    retrieve_full_text(dummy_ftjob.pk)

    assert dummy_ftjob.status == FullTextState.CANCELLED
    assert "cancelled" in dummy_ftjob.error_str.lower()
    dummy_ftjob.save.assert_called_once()
    dummy_ftjob.post.save.assert_called_once()


@patch("history4feed.h4fscripts.task_helper.h4f.get_full_text", side_effect=Exception("boom"))
@patch("history4feed.h4fscripts.task_helper.models.FulltextJob.objects.get")
def test_retrieve_full_text_exception(mock_get_job, mock_get_text, dummy_ftjob):
    mock_get_job.return_value = dummy_ftjob
    dummy_ftjob.is_cancelled.return_value = False

    retrieve_full_text(dummy_ftjob.pk)

    assert dummy_ftjob.status == FullTextState.FAILED
    assert dummy_ftjob.error_str == "boom"
    dummy_ftjob.save.assert_called_once()
    dummy_ftjob.post.save.assert_called_once()




import os
import time
from types import SimpleNamespace
import unittest, pytest
from urllib.parse import urljoin

from tests.utils import remove_unknown_keys, wait_for_jobs

base_url = os.environ["SERVICE_BASE_URL"]
import requests

@pytest.mark.parametrize(
    ["post_id", "should_fail"],
    [
        ["9c04d319-a949-52df-bcb6-5a73a1458fe5", True], #post does not exist
        ["4aa844cb-18e6-58cc-bed1-4c22abf3b977", False],
        ["4aa844cb-18e6-58cc-bed1-4c22abf3b977", True], #post already deleted
    ]
)
def test_delete_post(post_id, should_fail):
    post_url = urljoin(base_url, f"api/v1/posts/{post_id}/")
    delete_resp = requests.delete(post_url)

    if should_fail:
        assert delete_resp.status_code == 404, f"delete post request expected to fail: {delete_resp.text}"
        return
    assert delete_resp.status_code == 204, f"unexpected status, body: {delete_resp.text}"


    get_resp = requests.get(post_url)
    assert get_resp.status_code == 404, f"post should already be deleted"



"""
Django settings for history4feed project.

Generated by 'django-admin startproject' using Django 5.0.6.

For more information on this file, see
https://docs.djangoproject.com/en/5.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/5.0/ref/settings/
"""

import os
from pathlib import Path
from dotenv import load_dotenv
from datetime import UTC, datetime
from textwrap import dedent

load_dotenv()

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('DJANGO_SECRET', "insecure_django_secret")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.getenv('DJANGO_DEBUG', False)

ALLOWED_HOSTS = os.getenv('DJANGO_ALLOWED_HOSTS', "localhost 127.0.0.1 [::1]").split()

CELERY_BROKER_URL = os.environ["CELERY_BROKER_URL"]

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.redis.RedisCache',
        'LOCATION': CELERY_BROKER_URL,  # Use the appropriate Redis server URL
        'OPTIONS': {
            # 'CLIENT_CLASS': 'django.core.cache.backends.redis.RedisCacheClient',
        }
    }
}

#CORS_ALLOW_ALL_ORIGINS = os.environ.get('DJANGO_CORS_ALLOW_ALL_ORIGINS', True)
#CORS_ALLOWED_ORIGINS = [os.environ.get('DJANGO_CORS_ALLOWED_ORIGINS', "http://127.0.0.1:8002")]

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'drf_spectacular',
    'django.contrib.postgres',
    'history4feed.app',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'whitenoise.middleware.WhiteNoiseMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'history4feed.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'history4feed.wsgi.application'


# Database
# https://docs.djangoproject.com/en/5.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.getenv('POSTGRES_DB'),      # Database name
        'USER': os.getenv('POSTGRES_USER'),          # Database user
        'PASSWORD': os.getenv('POSTGRES_PASSWORD'),  # Database password
        'HOST': os.getenv('POSTGRES_HOST'),          # PostgreSQL service name in Docker Compose
        'PORT': os.getenv('POSTGRES_PORT'),  # PostgreSQL default port
    },
    'sqlite': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    },
}


# Password validation
# https://docs.djangoproject.com/en/5.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/5.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.0/howto/static-files/

STATIC_URL = 'static/'
STATIC_ROOT = BASE_DIR / "staticfiles"

# Default primary key field type
# https://docs.djangoproject.com/en/5.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'


REST_FRAMEWORK = {
    # YOUR SETTINGS
    'DEFAULT_SCHEMA_CLASS': 'history4feed.app.autoschema.H4FSchema',
    'DEFAULT_FILTER_BACKENDS': ['django_filters.rest_framework.DjangoFilterBackend'],
    'DEFAULT_AUTHENTICATION_CLASSES': [],
    'EXCEPTION_HANDLER': "dogesec_commons.utils.custom_exception_handler",
    'DATETIME_FORMAT': '%Y-%m-%dT%H:%M:%SZ',
}

SPECTACULAR_SETTINGS = {
    'TITLE': "history4feed API",
    'DESCRIPTION': dedent("""
        history4feed can be used to create a complete history for a blog and output it as an RSS feed.
    """),
    'VERSION': '1.0.0',
    'CONTACT': {
        'email': 'noreply@dogesec.com',
        'url': 'https://github.com/muchdogesec/history4feed',
    },
    'TAGS': [
        {
            "name": "Feeds",
            "description": "Subscribe and retrieve Feeds"
        },
        {
            "name": "Posts",
            "description": "Retrieve Posts in Feeds"
        },
        {
            "name": "Jobs",
            "description": "Check the status of data retrieval from Feeds"
        },
    ],

}

DEFAULT_PAGE_SIZE = int(os.getenv("DEFAULT_PAGE_SIZE", 50))
MAXIMUM_PAGE_SIZE = int(os.getenv("MAX_PAGE_SIZE", 50))

HISTORY4FEED_SETTINGS = {
    'WAYBACK_SLEEP_SECONDS': int(os.getenv("WAYBACK_SLEEP_SECONDS", 20)),
    'EARLIEST_SEARCH_DATE': datetime.strptime(os.environ.get("EARLIEST_SEARCH_DATE", "2024-01-01T00:00:00Z"), "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=UTC),
    'REQUEST_RETRY_COUNT': int(os.getenv("REQUEST_RETRY_COUNT", 3)),
}


from datetime import datetime, timezone
from xml.dom.minidom import Document, Element


def createTextElement(document: Document, tagName, text):
    el = document.createElement(tagName)
    txtNode = document.createTextNode(text or "")
    el.appendChild(txtNode)
    return el

def createCDataElement(document: Document, tagName, text):
    el = document.createElement(tagName)
    txtNode = document.createCDATASection(text or "")
    el.appendChild(txtNode)    
    return el

def createRSSHeader(title, description,  url, last_build_date=None):
    last_build_date = last_build_date or datetime.now(timezone.utc)
    d = Document()
    rss = d.createElement("rss")
    d.appendChild(rss)
    rss.setAttribute("version", "2.0")
    channel = d.createElement("channel")
    rss.appendChild(channel)
    channel.appendChild(createTextElement(d, "title", title))
    channel.appendChild(createTextElement(d, "description", description))
    channel.appendChild(createTextElement(d, "link", url))
    channel.appendChild(createTextElement(d, "lastBuildDate", last_build_date.isoformat()))
    # channel.appendChild(createTextElement(d, "generator", LINK_TO_SELF))
    return d, channel


def getText(nodelist: list[Element]):
    if not nodelist:
        return ''
    if not isinstance(nodelist, list):
        nodelist = nodelist.childNodes
    rc = []
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE or node.nodeType == node.CDATA_SECTION_NODE:
            rc.append(node.data)
    return ''.join(rc)

def getFirstElementByTag(node, tag):
    if not node:
        return None
    elems = node.getElementsByTagName(tag)
    return (elems or None) and elems[0]

def getFirstChildByTag(node: Element, tag):
    child = None
    for c in node.childNodes:
        if c.nodeName == tag:
            child = c
            break
    return child


def getAtomLink(node: Element, rel='self'):
    links = [child for child in node.childNodes if child.nodeType == child.ELEMENT_NODE and child.tagName in ['link', 'atom:link']]

    link = links[0]
    for l in links:
        r = l.attributes.get('rel')
        if r and r.value == rel:
            link = l
            break
    return link.attributes['href'].value


import json
import schemathesis, schemathesis.schemas
from schemathesis.specs.openapi.schemas import BaseOpenAPISchema
from schemathesis import Case
from schemathesis.transports.responses import GenericResponse

@schemathesis.hook
def after_load_schema(
    context: schemathesis.hooks.HookContext,
    schema: BaseOpenAPISchema,
) -> None:
    
    schema.add_link(
        source=schema["/api/v1/jobs/"]['GET'],
        target=schema["/api/v1/jobs/{job_id}/"]['GET'],
        status_code=200,
        parameters={"path.job_id": '$response.body#/jobs/0/id'}
    )
    for method in ['GET', 'PATCH', 'DELETE']:
        schema.add_link(
            source=schema['/api/v1/feeds/']['GET'],
            target=schema['/api/v1/feeds/{feed_id}/'][method],
            status_code=200,
            parameters={"path.feed_id": "$response.body#/feeds/0/id"}
        )

    for method in ['GET', 'PATCH', 'DELETE']:
        schema.add_link(
            source=schema['/api/v1/posts/']['GET'],
            target=schema['/api/v1/posts/{post_id}/'][method],
            status_code=200,
            parameters={"path.post_id": "$response.body#/posts/0/id"}
        )


import os
import time
from types import SimpleNamespace
import unittest, pytest
from urllib.parse import urljoin

from tests.utils import remove_unknown_keys, wait_for_jobs

base_url = os.environ["SERVICE_BASE_URL"]
import requests


DATA = [
    {
        "id": "d1d96b71-c687-50db-9d2b-d0092d1d163a",
        "feed_type": "rss",
        "include_remote_blogs": False,
        "url": "https://muchdogesec.github.io/fakeblog123/feeds/rss-feed-encoded.xml",
    },
    {
        "id": "cb0ba709-b841-521a-a3f2-5e1429f4d366",
        "feed_type": "atom",
        "pretty_url": "https://muchdogesec.github.io/fakeblog123/",
        "title": "Custom Title",
        "description": "Custom description",
        "include_remote_blogs": False,
        "url": "https://muchdogesec.github.io/fakeblog123/feeds/atom-feed-decoded.xml",
    },
    {
        "id": "121e5557-7277-5aa3-945d-e466c6bf92d5",
        "title": "Custom Title 2",
        "feed_type": "atom",
        "include_remote_blogs": False,
        "url": "https://muchdogesec.github.io/fakeblog123/feeds/atom-feed-cdata.xml",
    },
    {
        "id": "8f89731d-b9de-5931-9182-5460af59ca84",
        "description": "Custom description 2",
        "feed_type": "rss",
        "include_remote_blogs": False,
        "url": "https://muchdogesec.github.io/fakeblog123/feeds/rss-feed-decoded.xml",
    },
    {
        "id": "9c04d319-a949-52df-bcb6-5a73a1458fe5",
        "feed_type": "atom",
        "include_remote_blogs": False,
        "url": "https://muchdogesec.github.io/fakeblog123/feeds/atom-feed-decoded-partial.xml",
    },
    {
        "id": "d63dad15-8e23-57eb-80f7-715cedf85f33", # not passed in request
        "feed_type": "skeleton", # not passed in request
        "pretty_url": "https://muchdogesec.github.io/fakeblog123/about/",
        "url": "https://muchdogesec.github.io/fakeblog123/",
        "title": "Skeleton custom Title",
        "description": "Skeleton custom description"
    }
]

def all_blog_parameters():
    return [
        pytest.param(k["url"], k, k.get("should_fail", False))
            for k in DATA
    ]

@pytest.mark.parametrize(
        ["url", "blog_data", "should_fail"],
        all_blog_parameters(),
)
def test_add_blog(url, blog_data: dict, should_fail):
    payload = remove_unknown_keys(blog_data, ["pretty_url", "title", "description", "include_remote_blogs", "url"])

    endpoint = urljoin(base_url, "api/v1/feeds/")

    if blog_data["feed_type"] == "skeleton":
        post_resp = requests.post(urljoin(endpoint, "skeleton/"), json=payload)
    else:
        post_resp = requests.post(endpoint, json=payload)

    if should_fail:
        assert not post_resp.ok, "add feed request expected to fail"
        return

    assert post_resp.status_code == 201, f"request failed: {post_resp.text}"
    post_resp_data = post_resp.json()
    job_id = post_resp_data.get("job_id")
    feed_id = post_resp_data["id"]
    if job_id:
        wait_for_jobs(job_id)

    feed_resp = requests.get(urljoin(base_url, f"api/v1/feeds/{feed_id}/"))
    resp_data = feed_resp.json()

    assert resp_data["id"] == blog_data["id"]

    if expected_pretty_url := blog_data.get("pretty_url"):
        assert resp_data["pretty_url"] == expected_pretty_url

    if expected_title := blog_data.get("title"):
        assert resp_data["title"] == expected_title

    if expected_description := blog_data.get("description"):
        assert resp_data["description"] == expected_description

    if expected_feed_type := blog_data.get("feed_type"):
        assert resp_data["feed_type"] == expected_feed_type

    if payload.get('use_search_index'):
        assert resp_data["feed_type"] == "search_index"



import re
from textwrap import dedent
from typing import Iterable
from urllib.parse import urljoin, urlparse
import uuid
from .settings import history4feed_server_settings
from django.db import models
from rest_framework import validators
from uuid import uuid4
from django.utils.text import slugify
import hyperlink
from django.db.models import Min, Max
from django.db.models import OuterRef, Subquery
from django.db.models import F
from django.utils import timezone
from django.db import transaction

POST_DESCRIPTION_MAX_LENGTH = 2 * 1024 * 1024 # 2MiB
FEED_DESCRIPTION_MAX_LENGTH = 10*1024 # 10KiB

class JobState(models.TextChoices):
    PENDING    = "pending"
    RUNNING    = "running"
    SUCCESS    = "success"
    CANCELLED  = "cancelled"
    FAILED     = "failed"

class FeedType(models.TextChoices):
    RSS = "rss"
    ATOM = "atom"
    SKELETON = "skeleton"
    SEARCH_INDEX = "search_index"

# Create your models here.

class SlugField(models.CharField):
    def get_prep_value(self, value):
        return slugify(str(value))

class Category(models.Model):
    name = SlugField(max_length=1000, primary_key=True)


def stix_id(url):
    return uuid.uuid5(uuid.UUID(str(history4feed_server_settings.HISTORY4FEED_NAMESPACE)), url)

def normalize_url(url):
    try:
        url = re.sub(r'(?<!:)/{2,}', '/', url)
        u = hyperlink.parse(url)
        if not u.scheme:
            raise validators.ValidationError('url must be a full path')
        return u.normalize(url).to_text()
    except Exception as e:
        raise validators.ValidationError(f"URL normalization failed")
    
AUTO_TITLE_TRAIL = "%^%*(%"
def title_as_string(value: str):
    if value.endswith(AUTO_TITLE_TRAIL):
        value = value[:-len(AUTO_TITLE_TRAIL)]
    return value

class Feed(models.Model):
    id = models.UUIDField(primary_key=True, help_text="UUID of feed generated by history4feed")
    title = models.CharField(max_length=1000, help_text="found in the <channel> of RSS output. Is always kept up to date with the latest feed import values for this property.")
    description = models.CharField(max_length=FEED_DESCRIPTION_MAX_LENGTH, help_text="found in the <channel> of RSS output. Is always kept up to date with the latest feed import values for this property.", null=True, default=None)
    url = models.URLField(max_length=1000, unique=True, help_text=dedent("""
        The URL of the RSS or ATOM feed

        Note this will be validated to ensure the feed is in the correct format.
    """), validators=[normalize_url])
    earliest_item_pubdate = models.DateTimeField(null=True, help_text="pubdate of earliest post")
    latest_item_pubdate = models.DateTimeField(null=True, help_text="pubdate of latest post")
    datetime_added = models.DateTimeField(auto_now_add=True, editable=False, help_text="date feed entry was added to database")
    datetime_modified = models.DateTimeField(default=None, null=True, help_text="date feed entry was edited in the database")
    feed_type = models.CharField(choices=FeedType.choices, max_length=12, null=False, editable=False, help_text="type of feed")
    pretty_url = models.URLField(max_length=1000, null=True, default=None)
    freshness = models.DateTimeField(null=True, default=None)

    def get_post_count(self):
        return self.posts.filter(deleted_manually=False).count()
    
    def save(self, *args, **kwargs) -> None:
        if not self.id:
            self.id = stix_id(self.url)
        self.earliest_item_pubdate, self.latest_item_pubdate = self.posts.aggregate(min=Min('pubdate'), max=Max('pubdate')).values()
        self.datetime_modified = self.datetime_modified or self.datetime_added
        return super().save(*args, **kwargs)
    
    def get_pretty_url(self):
        return self.pretty_url or self.url
    
    def set_title(self, title):
        if not self.title or self.title.endswith(AUTO_TITLE_TRAIL):
            self.title = title + AUTO_TITLE_TRAIL
    
    def set_description(self, description):
        if not self.description or self.description.endswith(AUTO_TITLE_TRAIL):
            self.description = description + AUTO_TITLE_TRAIL

class Job(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4, help_text="UUID of job")
    state = models.CharField(choices=JobState.choices, max_length=12, default=JobState.PENDING, null=False, help_text="state of the job")
    run_datetime = models.DateTimeField(auto_now_add=True, editable=False, help_text="time job was executed")
    earliest_item_requested = models.DateTimeField(null=True, help_text="shows the earliest time for posts requested. Useful for when jobs are run to see if the time range it runs across is expected")
    latest_item_requested = models.DateTimeField(null=True, help_text="shows the latest time for posts requested")
    feed = models.ForeignKey(Feed, on_delete=models.CASCADE)
    info = models.CharField(max_length=FEED_DESCRIPTION_MAX_LENGTH, help_text="contains a useful summary of the job (e.g. number of posts retrieved, errors logged)")
    include_remote_blogs = models.BooleanField(default=False)

    def urls(self):
        retval = {}
        ft_job: FulltextJob = None
        for ft_job in self.fulltext_jobs.all():
            retval[ft_job.status] = retval.get(ft_job.status, [])
            retval[ft_job.status].append(dict(url=ft_job.link, id=ft_job.post_id))
        return retval
    
    def should_skip_post(self, post_link: str):
        return (not self.include_remote_blogs) and urlparse(self.feed.url).hostname.split('.')[-2:] != urlparse(post_link).hostname.split('.')[-2:]
    
    def cancel(self):
        self.update_state(JobState.CANCELLED)
        self.save()
        return
    
    def is_cancelled(self):
        return self.state == JobState.CANCELLED
    
    @transaction.atomic
    def update_state(self, state):
        obj = self.__class__.objects.select_for_update().get(pk=self.pk)
        if obj.state not in [JobState.PENDING, JobState.RUNNING]:
            return obj.state
        obj.state = state
        obj.save()
        self.refresh_from_db()
        return obj.state
    


class FullTextState(models.TextChoices):
    RETRIEVED  = "retrieved"
    SKIPPED    = "skipped"
    CANCELLED  = "cancelled"
    FAILED     = "failed"
    RETRIEVING = "retrieving"

class Post(models.Model):
    id = models.UUIDField(primary_key=True, help_text="UUID of items generated by history4feed")
    datetime_added = models.DateTimeField(auto_now_add=True, editable=False)
    datetime_updated = models.DateTimeField(auto_now=True)
    title = models.CharField(max_length=1000, help_text="found in the <item> element of feed output")
    description = models.CharField(max_length=POST_DESCRIPTION_MAX_LENGTH, blank=True, help_text="found in the <item> element of feed output")
    link = models.URLField(max_length=1000, help_text="link to full article. found in the <item> element of feed output", validators=[normalize_url])
    pubdate = models.DateTimeField(help_text="date of publication.")
    author = models.CharField(max_length=1000, help_text="author of the post", null=True, blank=True)
    categories = models.ManyToManyField(Category, related_name="posts", help_text="categories of the post", blank=True)
    feed = models.ForeignKey(Feed, on_delete=models.CASCADE, related_name="posts", help_text="feed id this item belongs too")
    is_full_text = models.BooleanField(default=False, help_text="if full text has been retrieved")
    content_type = models.CharField(default="plain/text", max_length=200, help_text="content type of the description")
    added_manually = models.BooleanField(default=False)
    deleted_manually = models.BooleanField(default=False, help_text="this post is hidden from user")

    class  Meta:
        constraints = [
            models.UniqueConstraint(fields=["link", "feed"], name="unique_link_by_feed"),
        ]

    def add_categories(self, categories):
        categories = categories or []
        categories = [Category.objects.get_or_create(name=name)[0] for name in categories]
        self.categories.set(categories)


    def save(self, *args, **kwargs) -> None:
        if not self.id:
            pubdate = self.pubdate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            self.id = stix_id(f"{self.feed.id}+{self.link}+{pubdate}")
        return super().save(*args, **kwargs)
    
    @classmethod
    def visible_posts(cls):
        return cls.objects.filter(deleted_manually=False).annotate(
            last_job_id=Subquery(
                FulltextJob.objects.filter(
                    post_id=OuterRef('pk')
                ).order_by('-job__run_datetime')  # Ordering by publication date to get the latest book
                .values('job__id')[:1]  # We take only the first (most recent) book
            )
        )

class FulltextJob(models.Model):
    post = models.ForeignKey(Post, on_delete=models.SET_NULL, null=True, related_name="fulltext_jobs")
    job = models.ForeignKey(Job, related_name="fulltext_jobs", on_delete=models.CASCADE)
    status = models.CharField(max_length=15, choices=FullTextState.choices, default=FullTextState.RETRIEVING)
    error_str = models.CharField(max_length=1500, null=True, blank=True)
    link = models.CharField(max_length=1500)


        
    def is_cancelled(self):
        return self.job.state == JobState.CANCELLED


from datetime import datetime, timezone
import os
from typing import Any, Dict, get_type_hints
import uuid
from django.core.signals import setting_changed

from django.conf import settings
from rest_framework.settings import APISettings, perform_import, api_settings

H4F_DEFAULTS: dict[str, any] = {
    'SCRAPFLY_APIKEY': os.getenv('SCRAPFLY_APIKEY'),
    'WAYBACK_SLEEP_SECONDS': 20,
    'EARLIEST_SEARCH_DATE': datetime(2020, 1, 1, tzinfo=timezone.utc),
    'REQUEST_RETRY_COUNT': 3,
    'HISTORY4FEED_NAMESPACE': uuid.UUID("6c6e6448-04d4-42a3-9214-4f0f7d02694e"),
    "BRAVE_SEARCH_API_KEY": None
}

IMPORT_STRINGS = [
]

class History4FeedServerSettings(APISettings):
    SCRAPFLY_APIKEY: str
    WAYBACK_SLEEP_SECONDS: int
    EARLIEST_SEARCH_DATE: datetime
    REQUEST_RETRY_COUNT: int
    HISTORY4FEED_NAMESPACE : str|uuid.UUID
    BRAVE_SEARCH_API_KEY: str

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'HISTORY4FEED_SETTINGS', {})
        return self._user_settings

history4feed_server_settings = History4FeedServerSettings(
    user_settings=getattr(settings, 'HISTORY4FEED_SETTINGS', {}),  # type: ignore
    defaults=H4F_DEFAULTS,  # type: ignore
    import_strings=IMPORT_STRINGS,
)


def reload_api_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'HISTORY4FEED_SETTINGS':
        history4feed_server_settings.reload()


setting_changed.connect(reload_api_settings)





from unittest.mock import patch
import uuid
from django.core import validators
import pytest
from history4feed.app import models
from history4feed.app.settings import history4feed_server_settings
from rest_framework.validators import ValidationError


def test_stix_id():
    url = 'https://goo.gl'
    with patch('uuid.uuid5') as mock_uuid5:
        gen_id = models.stix_id(url)
        mock_uuid5.assert_called_once_with(uuid.UUID(str(history4feed_server_settings.HISTORY4FEED_NAMESPACE)), url)
        assert gen_id == mock_uuid5.return_value
    assert models.stix_id(url) == models.stix_id(url), "should be deterministic"

def test_normalize_url():
    assert models.normalize_url('https://example.com/net/') == 'https://example.com/net/'
    assert models.normalize_url('https://example.com//net//') == 'https://example.com/net/'
    assert models.normalize_url('http://example.com///net//a////') == 'http://example.com/net/a/'
    assert models.normalize_url('http://example.com///net') == 'http://example.com/net'
    assert models.normalize_url('http://example.com') == 'http://example.com/', "should add trailing slash when path is empty"

    with pytest.raises(ValidationError):
        models.normalize_url('kjasjkaskj')

@pytest.mark.django_db
def test_save_feed():

    orig_url = 'http://example.com///net//a////'
    feed = models.Feed.objects.create(title='my title', description='my description', url=orig_url)
    feed.save()

    assert feed.get_pretty_url() == orig_url
    assert feed.id == models.stix_id(feed.url)

def test_feed_get_pretty_url():
    orig_url = 'http://example.com'
    feed = models.Feed(title='my title', description='my description', url=orig_url)
    assert feed.get_pretty_url() == feed.url
    feed.pretty_url = 'https://example.net/pretty-url/'
    assert feed.get_pretty_url() == feed.pretty_url
